#!/usr/bin/env python

# COPYRIGHT NOTICE

"""
* Copyright (c) 2008, University of Brescia, Brescia, IT
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Brescia nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY UNIVERSITY OF BRESCIA ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL UNIVERSITY OF BRESCIA BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

#  ABOUT

"""
 *
 *           m o d f c a p . p y
 *
 * This module handle the TCP flow tracking and the flows lookup table.
 *
 * Author -
 *
 *      Francesco Gringoli
 *      francesco.gringoli@ing.unibs.it
 *      University of Brescia
 *      1/3/2008
 *
 * Status -
 *      Stable
 *
"""

#-------------------------------------------------------------------------------
# This module handles I/O concerning pcap

#-------------------------------------------------------------------------------
# TODO
#
# - clean the code
# - add parts for anomaly detection implemented in Wien
# - add knowledge of different layer 2 format (now only ethernet,
#   ethertype==IP is supported; e.g. add vlan, hdlc ...)
# - add a brief introduction on status numbers

#-------------------------------------------------------------------------------
# import section

import pcap, dpkt, dnet
import sys
import flow
import options
import modules
import util

#-------------------------------------------------------------------------------
# This class implements a simple look-up table for flow handling
#
# Method __init__ initializes the dictionary below this object
# Method 'insertflow' inserts the passed flow inside the db
# Method 'removeflow' removes the passed flow from the db
# Method 'lookupflow' searches for a flow inside the db
# Method 'countflows' counts how many flow definitions are into db
#

class TrackDB:

    def __init__( self ):
        self.db = {}
        self.flowno = 0

    def insertflow( self, f ):

        if util.getverbosity() >= 2:
            util.log( f.flow2str() + " inserted at " + str(f.getcurtime()) )

        if not self.db.has_key( f.srcadd ):
            self.db[f.srcadd] = {}
        if not self.db[f.srcadd].has_key( f.dstadd ):
            self.db[f.srcadd][f.dstadd] = {}
        if not self.db[f.srcadd][f.dstadd].has_key( f.srcport ):
            self.db[f.srcadd][f.dstadd][f.srcport] = {}
        if not self.db[f.srcadd][f.dstadd][f.srcport].has_key( f.dstport ):
            self.db[f.srcadd][f.dstadd][f.srcport][f.dstport] = f
	
        # we have one more flow
        self.flowno = self.flowno + 1

        # mark as not orphan
        f.setorphan(False)

    def removeflow( self, f ):
        # although we should never be called to remove a flow that does not exist
        # this could happen! Hence handle this case with a try. Attention:
        # we can decrease the flow number only after we are sure the flow was still there
        try:
        
            dstaddlist_ = self.db[f.srcadd]
            srcportlist_ = dstaddlist_[f.dstadd]
            dstportlist_ = srcportlist_[f.srcport]

            # remove the flow
            del dstportlist_[f.dstport]

            # mark as orphan
            f.setorphan(True)

            if util.getverbosity() >= 2:
                util.log( f.flow2str() + " removed" )

            # ok, if we got here the flow was in the db, decrease flow number
            self.flowno = self.flowno - 1
 
            # now check how was the flow nested and remove empty list
            if len( dstportlist_ ) > 0:
                return
            del srcportlist_[f.srcport]

            if len( srcportlist_ ) > 0:
                return
            del dstaddlist_[f.dstadd]

            if len( dstaddlist_ ) > 0:
                return
            del self.db[f.srcadd]

        except:
            util.err( "Attempt to remove " + f.flow2str() + " lead to a miss" )

    def lookupflow(self,ippkt_):
        
        # if ippkt_ belongs to an already defined flow, return this flow
        if self.db.has_key( ippkt_.src ) and self.db[ippkt_.src].has_key( ippkt_.dst ):
            if self.db[ippkt_.src][ippkt_.dst].has_key( ippkt_.data.sport ) and \
                   self.db[ippkt_.src][ippkt_.dst][ippkt_.data.sport].has_key( ippkt_.data.dport ):
                f = self.db[ippkt_.src][ippkt_.dst][ippkt_.data.sport][ippkt_.data.dport]
                f.setdir( "fclient" )
                return f
            
        if self.db.has_key( ippkt_.dst ) and self.db[ippkt_.dst].has_key( ippkt_.src ):
            if self.db[ippkt_.dst][ippkt_.src].has_key( ippkt_.data.dport ) and \
                   self.db[ippkt_.dst][ippkt_.src][ippkt_.data.dport].has_key( ippkt_.data.sport ):
                f = self.db[ippkt_.dst][ippkt_.src][ippkt_.data.dport][ippkt_.data.sport]
                f.setdir( "fserver" )
                return f
            
        return None

    def countflows( self ):
        return self.flowno

    def realcountflows( self ):            
        flows = 0
        
        # count all flows in the DB
        for srcadd_ in self.db:
            for dstadd_ in self.db[srcadd_]:
                for srcport_ in self.db[srcadd_][dstadd_]:
                    for dstport_ in self.db[srcadd_][dstadd_][srcport_]:
                        flows = flows + 1
        return flows

    def __iter__( self ):
        self.sadditerator = None
        self.currentflow = None
        return self

    def next( self ):
        if self.currentflow != None and self.currentflow.isfinished():
            self.removeflow( self.currentflow )
            self.currentflow = None

        if self.sadditerator == None:
            try:
                self.sadditerator = self.db.keys()
                self.sadditer = 0
                self.dadditerator = self.db[self.sadditerator[0]].keys()
                self.dadditer = 0
                self.sprtiterator = self.db[self.sadditerator[0]][self.dadditerator[0]].keys()
                self.sprtiter = 0
                self.dprtiterator = self.db[self.sadditerator[0]][self.dadditerator[0]][self.sprtiterator[0]].keys()
                self.dprtiter = 0

            except:
                if util.getverbosity() >= 2:
                    util.err("db iterator end")
                raise StopIteration

        else:
            self.dprtiter = self.dprtiter + 1
            if self.dprtiter == len ( self.dprtiterator ):
                self.sprtiter = self.sprtiter + 1
                if self.sprtiter == len ( self.sprtiterator ):
                    self.dadditer = self.dadditer + 1
                    if self.dadditer == len ( self.dadditerator ):
                        self.sadditer = self.sadditer + 1
                        if self.sadditer == len ( self.sadditerator ):
                            self.sadditerator = None
                            self.dadditerator = None
                            self.sprtiterator = None
                            self.dprtiterator = None
                            raise StopIteration
                        
                        self.dadditerator = self.db[self.sadditerator[self.sadditer]].keys()
                        self.dadditer = 0

                    self.sprtiterator = self.db[self.sadditerator[self.sadditer]][self.dadditerator[self.dadditer]].keys()
                    self.sprtiter = 0

                self.dprtiterator = self.db[self.sadditerator[self.sadditer]][self.dadditerator[self.dadditer]][self.sprtiterator[self.sprtiter]].keys()
                self.dprtiter = 0

        self.currentflow = self.db[self.sadditerator[self.sadditer]][self.dadditerator[self.dadditer]][self.sprtiterator[self.sprtiter]][self.dprtiterator[self.dprtiter]]
            
        return self.currentflow


#-------------------------------------------------------------------------------
# This class implements the modfcap engine
#
# Method 'modinfo' returns the message help for this module
# Method 'default' returns the dictionary with default config for this module
# Method __init__ initializes this module and creates the db
# Method 'setup' passes the supplied configuration to this module
# Method 'next' iterates over all packets
# Method 'countflows' returns how many flows are still inside the db
#

class modfcap(modules.template):
    def modinfo( self ):
        info = ( "fcap module, call with -m fcap:p1=v1:p2=v2:... with pi and vi given by\n" +
                 "  [if=input_file_name]\n" +
                 "  [snap=snap_len]\n" +
                 "  [promisc=0|1]")
        return info

    def default( self ):
        return {"if":"","snap":0,"promisc":0}
    
    def __init__( self ):
        # call the base class setup function
        modules.template.setup( self )

        # initialize parameters
        self.lastflow = None
        
        # allocates a new DB to track flows
        self.conntrack=TrackDB()

    def setup( self, config ):
        self.input = config['if']
        self.snaplen = config['snap']
        self.promisc = config['promisc']
	self.fileno = 0
        
        # if requested allocates a new packet capture engine and set 'I'
        # input capability
        if len(self.input)>0:
            try:
                self.pc = pcap.pcap( self.input, self.snaplen, self.promisc, False )
                self.capabilities.append( 'I' )
            except:
                #self.errormsg = "Error opening the capture stream"
                util.err("Error opening the capture stream")

    def nextfile( self ):
        self.fileno += 1
        try:
            util.err( "Trying next file " + self.input + str(self.fileno) )
            self.pc = pcap.pcap( self.input + str(self.fileno), self.snaplen, self.promisc, False )
        except:
            util.err("Error opening the capture stream")
            raise StopIteration
	
	util.err( "Ok opening: " + self.input + str(self.fileno) )
    	

    def getdb( self ):
        return self.conntrack
    
    def __iter__( self ):
        return self

    def next( self ):
        while True:
            try:

                # check if the last observed flow should be inserted into database
                # or on the contrary it should be removed because we must skip it or it is finished
                if self.lastflow != None:
                    if self.lastflow.isorphan() and not self.lastflow.skip():
                        self.conntrack.insertflow( self.lastflow )
                    elif not self.lastflow.isorphan() and (self.lastflow.skip() or self.lastflow.isfinished()):
                        self.conntrack.removeflow( self.lastflow )

                # get a new packet
                retpkt = self.pc.next()

                if retpkt == None:
                    util.err( "pcap interface returned none" )
		    self.nextfile()
                    continue

                (ti, pkt) = retpkt

                # ADD HERE code to handle different layer 2 formats
                # now we handle only ethernet with standard ethertype ip
                if dpkt.ethernet.Ethernet( pkt ).type != 2048:
                    continue
                ippkt = dpkt.ethernet.Ethernet( pkt ).data

                # if packet is empty continue
                if ippkt.len == 0:
                    continue

                # handle only TCP and UDP
                if ippkt.p != 6 and ippkt.p != 17:
                    continue
                
                # try to identify this flow: if unknown, flow is empty, create a new flow
                f = self.conntrack.lookupflow( ippkt )
                if not f:
                    f = flow.flow( ippkt )

                # store pkt and time into the flow
                f.setcurl2pkt( pkt )
                f.setcurtime( ti )

		# store the flow so that we will analyze it during the next loop
		self.lastflow = f

                return f
        
            except: # StopIteration:
                util.err("end of pcap loop")
                self.nextfile()
		continue

    def countflows(self):
        return self.conntrack.countflows()

    def realcountflows(self):
        return self.conntrack.realcountflows()
