#!/usr/bin/env python

# COPYRIGHT NOTICE

"""
* Copyright (c) 2008, University of Brescia, Brescia, IT
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Brescia nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY UNIVERSITY OF BRESCIA ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL UNIVERSITY OF BRESCIA BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

#  ABOUT

"""
 *
 *           m o d g t r u t h . p y
 *
 * This module handle the ground truth matching of TCP flows.
 *
 * Author -
 *
 *      Francesco Gringoli
 *      francesco.gringoli@ing.unibs.it
 *      University of Brescia
 *      21/11/2008
 *
 * Status -
 *      Stable
 * 
"""

#-------------------------------------------------------------------------------
# This module implements the match engine toward a csv file

#-------------------------------------------------------------------------------
# TODO
#
#-------------------------------------------------------------------------------
# import section

import sys
import flow
import options
import modules
import re
import dpkt
import string
import util
import math

#-------------------------------------------------------------------------------
# This class defines the csv quering module
#
# Method doprocess: process the TCP payload searching for known protocol patterns

class modgtruth(modules.template):
    def modinfo( self ):
        info = ( "gtruth module, call with -m gtruth:if=csvfilename\n" +
                 "\n" +
                 "  This module will mark a flow with the application name gathered from the csv repository\n" )
        return info

    def default( self ):
        return {"if":""}
    
    def __init__( self ):
        
        # call the base class setup function
        modules.template.setup( self )

    def loadudp( self, fname ):

        file = open(fname, 'r')

        timedb = {}
        nowdb = {}

        nullhost = '0.0.0.0'

        for line in file:
            fields = line.split('%%%%')

            # use try to extract the port number so that we also skip the (if present) table header line
            now = fields[0]
            host = fields[1].strip()
            src  = fields[2].strip()
            dst  = fields[3].strip()
            try:
                srcp = int(fields[4].strip())
                dstp = int(fields[5].strip())
            except:
                continue
            appn = fields[6].strip()
            prot = fields[7].strip()
            logt = fields[8].strip()

            # consider just the first name... TO BE CORRECTED!
            if ':' in appn:
                names = appn.split(':')
                appn = names[0]

            # if the logger was restarted or just started for the first time remove all applications and related sockets
            if appn == 'restart' and \
               src == nullhost and dst == nullhost and \
               srcp == 0 and dstp == 0 and prot == 'UDP':

                nowdb[host] = {}
                continue

            # if this is a UDP socket allocation event...
            if src == nullhost and dst == nullhost and \
               srcp != 0 and dstp == 0 and prot == 'UDP':

                # the following line is to have logger that didn't wrote restart... old client
                if host not in nowdb:
                    nowdb[host] = {}

                # fetch the current db for this host
                if host in nowdb:

                    # host in db, fetch its applications list
                    hostapps = nowdb[host]

                    # if socket creation event:
                    if logt == 'created':

                        # check if this port exists
                        found = False
                        for app in hostapps:
                            if srcp in hostapps[app]:
                                found = True
                                break

                        if found == True:

                            # this port was already allocaed
                            util.err("duplicated port info for host %s port %d app %s" % (host, srcp, appn))

                        else:

                            # add this port to the application ports list
                            if appn not in hostapps:
                                hostapps[appn] = [ srcp ]
                            else:
                                hostapps[appn].append( srcp )

                    # if socket destruction event
                    elif logt == 'destroyed':

                        # check first that this port is hold by the same application
                        if appn not in hostapps or srcp not in hostapps[appn]:
                            util.err("port %d to be removed for host %s app %s but it does not exist" % (srcp, host, appn))
                        else:
                            hostapps[appn].remove( srcp )
                            if len(hostapps[appn]) == 0:
                                del hostapps[appn]

                    # cumulative db refresh: add this host if needed
                    if host not in timedb:
                        timedb[host] = []

                    # prepare a new element and add it
                    nel = {}
                    for app in hostapps:
                        nel[app] = hostapps[app][:]

                    if len(timedb[host]) and timedb[host][-1][0] == now:
                        timedb[host][-1] = ( now, nel )
                    else:
                        timedb[host].append( (now, nel ) )

        file.close()        
        return timedb

    def load( self, ts, te):

        # remove old db
        del self.gtdb
        self.gtdb = []

        # use this regexp to fetch the timestamp
        p = re.compile('(\d+)')

        # open the csv file
        try:
            f = open(self.filename, 'r')
        except IOError:
            util.err("modgtruth: csv file not found\n")
            return None

        # load data
        relem = 0

        # clean limits
        self.leftlim = 0
        self.rightlim = 0

        # save the requested limits
        self.reqleftlim = ts
        self.reqrightlim = te

        # widen the window so that we are sure that the requested limits have been requested
        ts = ts - 5
        te = te + 5

        if util.getverbosity() >= 2:
            util.log("  LOAD: window is widen at ts = %d, te = %d" % (ts, te))
            util.log("  LOAD: reqleftlim = %d, reqrightlim %d" % (self.reqleftlim, self.reqrightlim))

        try:
            for line in f:
                m = re.match(p, line)

                if m and int(m.group(1)) >= ts and int(m.group(1)) <= te:
                    if self.leftlim == 0:
                        self.leftlim = int(m.group(1))
                    self.rightlim = int(m.group(1))
                    fields = line.split('%%%%')
                    str2build = m.group(1)
                    for field in fields[1:]:
                        str2build = str2build + ':' + field.strip().replace(":", "$$$$")
                    self.gtdb.append(str2build)
                    relem = relem + 1

            f.close()

            if relem == 0:
                util.err("modgtruth: given limits do not return any flow")

            if util.getverbosity() >= 2:
                util.log("  LOAD: leftlim = %d, rightlim = %d" % (self.leftlim, self.rightlim))

            return relem

        except IOError:
            util.err("modgtruth: error reading file")
            f.close()
            return None

    def  findudpflow( self, ttime, host ):

        if host not in self.udpdb:
            return None

        li = 0
        ri = len(self.udpdb[host]) - 1

        fi = self.findudpflowrec(li, ri, ttime, host)
        
        if ttime < int( self.udpdb[host][fi][0] ):
            fi = fi - 1

        if fi < 0:
            return None

        if ttime > int( self.udpdb[host][fi][0] ):
            return fi

        return None
    
    def findudpflowrec(self, li, ri, ttime, host):

        # if these conditions hold we have finished searching
        if li == ri:
            return li
        if li == ri -1:
            return li

        # check in the middle
        mi = int( (li + ri) / 2)

        # take the element in the middle (time part)
        te = self.udpdb[host][mi][0]
        if( int(ttime) > int(te)):
            return self.findudpflowrec(mi, ri, ttime, host)
        else:
            return self.findudpflowrec(li, mi, ttime, host)
        
    def findflow( self, ttime ):
        li = 0
        ri = len(self.gtdb) - 1

        return self.findflowrec(li, ri, ttime)

    def findflowrec(self, li, ri, ttime):

        # if these conditions hold we have finished searching
        if li == ri:
            return li
        if li == ri - 1:
            return li

        # check in the middle
        mi = int( (li + ri) / 2)

        # take the element in the middle
        te = self.gtdb[mi]

        fields = te.split(':')

        # recurse depending on the found timestamp
        if( int(ttime) > int(fields[0])):
            return self.findflowrec(mi, ri, ttime)
        else:
            return self.findflowrec(li, mi, ttime)

    def setup( self, config ):

        # get configuration
        self.filename = config['if']

        # if no filename given exit
        if len(self.filename) == 0:
            util.err("modgtruth: no filename given")
            util.exit(-1)

        # use these variables to store the left-right limits of the loaded csv
        self.leftlim = 0
        self.rightlim = 0

        # use these variables to store the left-right limits of the requested csv (that can be really
        # different from limits above, usually wider!)
        self.reqleftlim = 0
        self.reqrightlim = 0

        # db is still empty
        self.gtdb = None

        # fill the udp db
        self.udpdb = self.loadudp(config['if'])

        #for host in self.udpdb:
        #    print host
        #
        #    timedbhost = self.udpdb[host]
        #
        #    for el in timedbhost:
        #        print el

        # append processing capability inside the list
        self.capabilities.append( 'P' )

    def flowgetdata( self, flow ):

        return self.flowattr( flow )

    def attrdefault( self ):

        # return default attribute for flow initialization
        attrib = dict()
        attrib["appname"] = "";
        attrib["time0"] = 0
        attrib["time"] = 0
        attrib["lasttime"] = 0
        attrib["freshlasttime"] = True
        attrib["fclient"] = 0
        attrib["fserver"] = 0
        attrib["l4proto"] = 'l4unknown'

        data = dict()
        
        return (attrib, data)
    
    def doprocess( self, flow ):

        # should we skip this flow?
        if flow.skip():
            return

        if flow.haspacket():
            # get module attribute for this flow
            (attrib, data) = self.flowattr( flow )

            # mark this time as the last we saw a packet
            # if this is a RST or FIN pkt (must be tcp...) then
            # don't put time in lasttime
            ttime = float( flow.getcurtime() )
            attrib["time"] = ttime
            attrib[flow.getdir()] = attrib[flow.getdir()] + flow.getpktl3len()

            # get packet and l4 proto
            tmppkt = flow.getcurl2pkt()
            tmpeth = dpkt.ethernet.Ethernet( tmppkt )
            pktproto = tmpeth.data.p

            # if this is tcp adjust lasttime excluding RST and shutdown procedure packets
            if pktproto == 6:
                attrib["l4proto"] = 'TCP'
                tmpflags = tmpeth.data.data.flags
                if tmpflags & 1 or tmpflags & 4:
                    attrib["freshlasttime"] = False

                if attrib["freshlasttime"] == True:
                    attrib["lasttime"] = ttime
            elif pktproto == 17:
                attrib["l4proto"] = 'UDP'
                attrib["lasttime"] = ttime
            else:
                util.err("Not allowed protocol, exit")
                util.exit(-1)

            # if we have never seen this flow search the associated app
            # we will be called for each TCP flow at the first SYN,
            # mark the flow so that we attempt resolution once per flow
            if len(attrib["appname"]) == 0:

                # mark time of this (first) packet and mark the flow
                # as resolution was ko
                attrib["time0"] = ttime
                attrib["appname"] = "NoName"
                (srcip, dstip, srcpt, dstpt ) = flow.flow2ascii()

                # if UDP let's try anonymous socket
                if pktproto == 17:
                    fi = self.findudpflow(ttime, srcip)
                    udphost = srcip
                    udpport = srcpt

                    if fi == None:
                        fi = self.findudpflow(ttime, dstip)
                        udphost = dstip
                        udpport = dstpt

                    # we found an interval for this host, check the port-app
                    if fi != None:
                        for app in self.udpdb[udphost][fi][1]:
                            if int(udpport) in self.udpdb[udphost][fi][1][app]:
                                attrib["appname"] = app
                                return

                # We get here if TCP (the block above is not executed) or UDP but the block above failed
                # if ttime is outside the limits of the old requested limits then load
                # a new piece of db
                if int(ttime) < self.reqleftlim or int(ttime) > self.reqrightlim:

                    if util.getverbosity() >= 2:
                        util.log(str(ttime) + " < " + str(self.reqleftlim) + " or " + str(ttime) + " > " + str(self.reqrightlim))

                    util.err("modgtruth: flow miss, loading db from csv")
                    ne = self.load(int(ttime) - 36000, int(ttime) + 36000)
                    if ne == None:
                        util.err("modgtruth: error reading csv")
                        util.exit(-1)

                    if ne == 0:
                        util.err("modgtruth: requested db from csv is empty")
                        return

                if util.getverbosity() >= 2:
                    util.log("after load we have leftlim = %d, rightlim = %d" % (self.leftlim, self.rightlim))
                    util.log("after load we have reqleftlim = %d, reqrightlim = %d" % (self.reqleftlim, self.reqrightlim))

                # if ttime is outside the limits of the loaded csv then return
                # do not compare exactly to limits... do some more useless cycles but
                # don't lose information! Widen to more than 1000s... prevents error due to
                # clock skews between pcap device and mysql host or delay due to gt in logging to mysql server
                if int(ttime) < (self.leftlim - 1001) or int(ttime) > (self.rightlim + 1001):

                    if util.getverbosity():
                        util.log("int(ttime) < (self.leftlim - 1001) or int(ttime) > (self.rightlim + 1001): skipping")

                    util.err("modgtruth: ttime outside limits of loaded csv and inside requested limits")
                    return

                # search for 100 seconds around the flow time
                li = self.findflow(int(ttime) - 100) - 1
                ei = self.findflow(int(ttime) + 100) + 1

                if li < 0:
                    li = 0

                if ei > len(self.gtdb) -1:
                    ei = len(self.gtdb) -1

                # loop over all the flows from li to ei
                for gtflow in self.gtdb[li:ei]:

                    #extract fields from gt flow
                    gtfields = gtflow.split(':')
                    gtrtime = gtfields[0]
                    gtuidip   = gtfields[1]
                    gtlocalip   = gtfields[2]
                    gtforeignip = gtfields[3]
                    gtlocalpt   = gtfields[4]
                    gtforeignpt = gtfields[5]
                    gtappnm = gtfields[6]
                    gtproto = gtfields[7]

                    # put here the code to check if srcip is local or foreigner. If local do nothing, otherwise
                    # swap src and destination (for both ip and port)

                    # TCP stuff
                    if pktproto == 6 and gtproto == 'TCP':
                        # forward check
                        if srcip == gtlocalip and dstip == gtforeignip and srcpt == gtlocalpt and dstpt == gtforeignpt:
                            attrib["appname"] = gtappnm
                            break

                        # reverse check
                        if dstip == gtlocalip and srcip == gtforeignip and dstpt == gtlocalpt and srcpt == gtforeignpt:
                            attrib["appname"] = gtappnm
                            break

                    # UDP stuff
                    if pktproto == 17 and gtproto == 'UDP':
                        # forward check
                        if srcip == gtlocalip and dstip == gtforeignip and srcpt == gtlocalpt and dstpt == gtforeignpt:
                            attrib["appname"] = gtappnm
                            break

                        # reverse check
                        if dstip == gtlocalip and srcip == gtforeignip and dstpt == gtlocalpt and srcpt == gtforeignpt:
                            attrib["appname"] = gtappnm
                            break

                        # sucking check 1
                        if (srcip == gtlocalip and srcpt == gtlocalpt) or (dstip == gtlocalip and dstpt == gtlocalpt):
                            attrib["appname"] = gtappnm
                            break

                        # sucking check 2
                        if (srcip == gtuidip and srcpt == gtlocalpt) or (dstip == gtuidip and dstpt == gtlocalpt):
                            attrib["appname"] = gtappnm
                            break
