#!/usr/bin/env python

# COPYRIGHT NOTICE

"""
* Copyright (c) 2008, University of Brescia, Brescia, IT
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Brescia nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY UNIVERSITY OF BRESCIA ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL UNIVERSITY OF BRESCIA BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

#  ABOUT

"""
 *
 *           o p t i o n s . p y
 *
 * options.py handles option parsing
 *
 * Author -
 *
 *      Francesco Gringoli
 *      francesco.gringoli@ing.unibs.it
 *      University of Brescia
 *      1/3/2008
 *
 * Status -
 *      Stable
 * 
"""

#-------------------------------------------------------------------------------
# This file handles
#
# - the configuration used by all python instances
# - limit definitions

#-------------------------------------------------------------------------------
# TODO
#
# - add comments to class Options variables
# - write code to allow Limits and Patterns override
# - recode handler for option -P: ask Manuel why a parameter can be specified

#-------------------------------------------------------------------------------
# import section

import sys
import getopt
import modules
import util

#-------------------------------------------------------------------------------
# This class defines an object to store the configuration given on the command
# line
#
# Method 'parse' parses the array with the command line
# Method 'usage' writes the usage instruction
#
# STATUS add comments

class options( object ):
    def __init__( self ):
        self.verbosity = 0
        self.progname = 'ipclass_traffic_suite'
        self.modules = []
        self.helprequest = []
        self.inputmodule = None
        self.outputmodules = []
        self.processmodules = []
        self.modnames = []
        
    def usage( self, msg = '' ):
        if len( msg ):
            util.log( "\n  " + msg + "\n\n" )
        else:
            util.log( "\n" +
                     "  Usage: " + self.progname + " [-v verbosity] [-h module_name] " +
                     "[-m module_configuration [-m module_configuration]]" +
                     "\n" +
                     "    -h this message" +
                     "\n" +
                     "    -v verbosity_level" +
                     "\n" +
                     "    -m module_name:config_string -- use config_string:=help to get help for this module" +
                     "\n\n" )

    def makemodconfig( self, default, config ):
        # 1. assignments not explicitly assigned by user but in default must be maintained
        newconfig = default

        # 2. user assignments that are out of default cannot be accepted
        for eqcoded in config:
            lrcoded = eqcoded.split( '=', 1 )
            lefthand = lrcoded[0]
            if len( lrcoded ) == 1:
                righthand = ''
            elif len( lrcoded ) == 2:
                righthand = lrcoded[1]
            else:
                return None
            if lefthand in default.keys():
                default[lefthand]=righthand
            else:
                return None

        return newconfig
        
    def parse( self, cmdline ):

	# if no parameters were given print usage and exit
	if len(cmdline) == 1:
	    self.usage()
	    sys.exit( -1 )

        try:
            opts, arguments = getopt.getopt( cmdline[1:], "v:m:h" )
        
        except getopt.GetoptError:
            self.usage( "Parameter not defined" )
            sys.exit( -1 )

        for o, a in opts:
            if o in ( "-v", ):
                util.setverbosity( a )

            if o in ( "-h", ):
                self.usage()
                sys.exit( -1 )
                
            if o in ("-m",):
                # this is a module request, find the name of the module, load it,
                # apply configuration. If error, report it and exit
                par = a.split(":")
                modname = par[0]

                # load the module
                newmod = modules.load( modname )
                if newmod == None:
                    self.usage( "The requested module '" + modname + "' can not be instantiated or does not exist" )
                    sys.exit( -1 )

                # set module name
                newmod.setname( modname )

                # check if help required
                if len(par) == 2:
                    if par[1] == "help":
                        self.usage( newmod.modinfo() )
                        sys.exit( -1 )
                    
                # prepare a dictionary with assignments comprising also default
                config = self.makemodconfig( newmod.default(), par[1:] )
                if config == None:
                    self.usage( "The configuration supplied for module '" + modname +
                                "' is not valid:\n\n" + newmod.modinfo() )
                    sys.exit( -1 )

                # setup the module with the prepared dictionary
                newmod.setup( config )
                if newmod.isonerror():
                    self.usage( "The configuration supplied for module '" + modname +
                                "' is not valid:\n\n" + newmod.errorstr() )
                    sys.exit( -1 )

                # check if this module is for input and verify it is the only one
                if newmod.isinput():
                    if self.inputmodule:
                        self.usage( "Only one module at time can be used as input" )
                        sys.exit( -1 )
                    self.inputmodule = newmod

                # check if this module is for output
                if newmod.isoutput():
                    self.outputmodules.append( newmod )
                    
                # check if this module is for processing
                if newmod.isprocess():
                    self.processmodules.append( newmod )

                # append this module
                self.modules.append( newmod )
                self.modnames.append( newmod.name() )

        # of no input module was specified print usage and exit
        if self.inputmodule is None:
            self.usage( "At least one input module must be specified" )
            sys.exit( -1 )
